#|
 Arxiu: roots.lisp
 Contingut: mdul per a resoldre equacions no linials
 Autor: Felip Alez Nadal
 ltima actualitaci: 9-VI-2006 |#


 #|BSD License

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. |#



(defpackage :ph_roots
  (:use :cl)
  (:export #:bisection #:newton-raphson #:steffensen #:regula-falsi)
  ( :documentation "This package contains functions to compute roots of non linear equations."))

(in-package :ph_roots)

(defun bisection ( f range &key (error_p 1e-3) (iterations 100))
  "Finds a root of a non linear equation using the bisection method."
  (unless (< (* (funcall f (first range)) (funcall f (first (rest range))) ) 0)
    (error "Non fited root. Please, use another range to fit the root.")
    
    )

  (let* (
	( a (first range))
	( b (first (rest range)))
	(tolerance (* (abs (- a b)) error_p))
	(iter 0)
	root
	fa
	fb
	fr
	step_error)
    (loop
     (setq root (/ (+ a b) 2))
     (setq fa (funcall f a))
     (setq fb (funcall f b))
     (setq fr (funcall f root))
     (cond
       ((< (abs fr) tolerance)
	(format t "Found root at ~D~%" root )
	(return-from bisection (coerce root 'double-float)))
       
       ((< (* fa fr) 0)
	(setq b root))
       
       (( < (* fb fr ) 0)
	(setq a root)))
     
     (setq step_error (abs (- a b)))
     (cond 
       ((<  step_error tolerance)
	(format t "Found root at ~D~%" root)
	(return-from bisection (coerce root 'double-float)))
       (t
	(incf iter)))
     (if (< iterations iter) (return)))))

(defun newton-raphson ( f df point &key (error_p 1e-3) (iterations 100))
  "Finds a root of a non linear equation using the Newton-Raphson method."
  (let
      (
       (root point)
       (iter 0)
       df0
       fr
       )
    (loop
     (setq df0 (funcall df root))
     (if (= df0 0)
	 (error "ZeroDivision Error. Too much near to a maximum or a minimum of f. Please, change the point to start the method."))
     (setq fr (funcall f root))
     (setq root (- root (/ fr df0)))
     (setq fr (funcall f root))
     (cond 
       ((< (abs fr) error_p)
	(format t "Found root at ~D~%" (coerce  root 'double-float))
	(return-from newton-raphson (coerce root 'double-float)))
       (t (incf iter)))
     (if (< iterations iter) (return)))))


(defun steffensen ( f  point &key (error_p 1e-3) (iterations 100))
  "Finds a root of a non linear equation using the Steffensen method."
  (let
      (
       (root point)
       (iter 0)
       df0
       fr
       )
    (loop
     (setq df0 (- (funcall f (+ root (funcall f root))) (funcall f root)))
     (if (= df0 0)
	 (error "ZeroDivision Error. Please, change the point to start the method."))
     (setq fr (funcall f root))
     (setq root (- root (/ (expt fr 2) df0)))
     (setq fr (funcall f root))
     (cond 
       ((< (abs fr) error_p)
	(format t "Found root at ~D~%" (coerce  root 'double-float))
	(return-from steffensen (coerce root 'double-float)))
       (t (incf iter)))
     (if (< iterations iter) (return)))))


(defun regula-falsi ( f range &key (error_p 1e-3) (iterations 100))
  "Finds a root of a non linear equation using the regula falsi method."
  (unless (< (* (funcall f (first range)) (funcall f (first (rest range))) ) 0)
    (error "Non fited root. Please, change the range.")
    
    )

  (let* (
	( a (first range))
	( b (first (rest range)))
	(tolerance (* (abs (- a b)) error_p))
	(iter 0)
	root
	fa
	fb
	fr
	step_error)
    (loop
    
     (setq fa (funcall f a))
     (setq fb (funcall f b))
     (if (= fa fb) (error "ZeroDivision error. Please, change the range.")) ;Divisi per zero
     
     (setq root (/ (- (* a fb) (* b fa)) (- fb fa )))
     (setq fr (funcall f root))
     
     (cond
       ((< (abs fr) tolerance)
	(format t "Found root at ~D~%" root )
	(return-from regula-falsi (coerce root 'double-float)))
       
       ((< (* fa fr) 0)
	(setq b root))
       
       (( < (* fb fr ) 0)
	(setq a root)))
     
     (setq step_error (abs (- a b)))
     (cond 
       ((<  step_error tolerance)
	(format t "Found root at ~D~%" root)
	(return-from regula-falsi (coerce root 'double-float)))
       (t
	(incf iter)))
     (if (< iterations iter) (return)))))
